/*! 
 * \file sugar.hpp
 * \brief Lazy header for fwdpp/sugar features
 */
#ifndef __FWDPP_SUGAR_SUGAR_HPP__
#define __FWDPP_SUGAR_SUGAR_HPP__

#include <fwdpp/sugar/GSLrng_t.hpp>
#include <fwdpp/sugar/infsites.hpp>
#include <fwdpp/sugar/metapop.hpp>
#include <fwdpp/sugar/multiloc.hpp>
#include <fwdpp/sugar/popgenmut.hpp>
#include <fwdpp/sugar/generalmut.hpp>
#include <fwdpp/sugar/serialization.hpp>
#include <fwdpp/sugar/singlepop.hpp>
#include <fwdpp/sugar/demography.hpp>
#include <fwdpp/sugar/sampling.hpp>
#include <fwdpp/sugar/add_mutation.hpp>
#include <fwdpp/sugar/matrix.hpp>
#endif
